/* acap.cc - acap-ish things
 * Rob Earhart
 * $Id: acap.cc,v 1.6 1998/03/09 00:57:50 rob Exp $
 */
/***********************************************************
        Copyright 1997 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of CMU not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.
******************************************************************/

#include <signal.h>
#include <signal.h>
#include <iostream.h>
#include <stdlib.h>
#include <netdb.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <syslog.h>
#include <unistd.h>
#include <sys/stat.h>
#include <pthread.h>
#include <fstream.h>
#include <fcntl.h>
#include "acap.h"
#include "conn.h"
#include "unixerror.h"
#include "dispatch.h"
#include "command.h"
#include "language.h"
#include "clientstr.h"
#include "trx.h"
#include "response.h"

Ptr<ACAP> ACAP::instance = NULL;

static void *TaskProcessor(void *ignore_me) {
  Dispatch::Instance()->Loop();
  return NULL;			// not that this should ever happen...
}

void ACAP::Shutdown() const {
  cerr << "Shutting server down" << endl << flush;
  Dispatch::Instance()->Shutdown();
  exit(0);
}

static void acap_sigshutdown(int sig, siginfo_t *foo, void *bar) {
  static bool shutting_down = false;
  if (shutting_down) {
    cerr << "Exiting due to " << _sys_siglist[sig] << " NOW" << endl << flush;
    syslog(LOG_ERR, "Exiting due to %s NOW\n", _sys_siglist[sig]);
    abort();
  }
  shutting_down = true;
  cerr << "Exiting due to " << _sys_siglist[sig] << endl << flush;
  syslog(LOG_ERR, "Exiting due to %s\n", _sys_siglist[sig]);
  ACAP::Instance()->Shutdown();
}

ACAP::ACAP(const char *path) {
  // We don't do anything with this anymore...
  // 'twas a nice idea while it lasted.
}

ACAP::~ACAP() {
}

int ACAP::Init(int argc, char **argv) {
  int c, sock, err = 0;
  const char *portname = "acap";
  struct sockaddr_in sin;
  bool rootset = false;

  char *acaproot = ACAP_ROOT;

  // Setting this and calling tzset seems to be the best
  // way of making sure that mktime() (used for Timestamp parsing)
  // will always be UCT, even if something calls tzset() again later.
  putenv("TZ=UCT");
  tzset();

  struct sigaction sa;
  sa.sa_handler = NULL;
  sa.sa_sigaction = &acap_sigshutdown;
  sigemptyset(&sa.sa_mask);
  sigaddset(&sa.sa_mask, SIGINT);
  sigaddset(&sa.sa_mask, SIGTERM);
  sa.sa_flags = SA_RESTART | SA_SIGINFO;
  if (sigaction(SIGINT, &sa, NULL))
    throw UnixError("Catching SIGINT");
  if (sigaction(SIGTERM, &sa, NULL))
    throw UnixError("Catching SIGTERM");

  sigset_t mask;
  sigemptyset(&mask);
  sigaddset(&mask, SIGPIPE);
  if (sigprocmask(SIG_BLOCK, &mask, NULL))
    throw UnixError("Blocking SIGPIPE");

  while ((c = getopt(argc, argv, "n:p:r:h")) != -1)
    switch (c) {
    case 'n':
      // Why can't nice() return a zero on success, like the man page says?
      errno = 0;
      nice(atoi(optarg));
      if (errno)
	throw UnixError("Illegal nice value");
      break;
    case 'p':
      portname = optarg;
      break;
    case 'r':
      acaproot = optarg;
      rootset = true;
      break;
    case 'h':
    case '?':
    default:
      err++;
      break;
    }
  if (err)
    throw Error("Usage", "acapd [-p port] [-r root-directory] [-n nice-value]");

  instance = new ACAP(acaproot);

  if (chdir(acaproot)) {
    cerr << "Unable to access data directory \"" << acaproot << '\"';
    if (! rootset)
      cerr << "; try using the -r switch to specify a different one";
    cerr << '.' << endl;
    throw UnixError("chdir");
  }

  openlog("acapd", LOG_PID | LOG_NDELAY
#ifdef sgi
          | LOG_PERROR
#endif
          , LOG_LOCAL0);

  // Set up processing threads...
  pthread_attr_t attr;
  pthread_attr_init(&attr);
  pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM);
  for (unsigned lupe = 0; lupe < 8; lupe++)
    pthread_create(NULL, &attr, &TaskProcessor, NULL);
  pthread_attr_destroy(&attr);

  // Now that we've got our task processing threads in place,
  // we can recover any pending transactions...
  Trx::Recover();

  // Now that we're all recovered, we can construct the socket,
  // and start listening for requests.

  // Set up the socket name...
  memset(&sin, 0L, sizeof(struct sockaddr_in));
  sin.sin_family = AF_INET;
  // See if the port works as a number...
  if (! (sin.sin_port = htons(atoi(portname)))) {
    // no?  okay, let's try getservbyname...
    struct servent *se = getservbyname(portname, "tcp");
    if (se)
      sin.sin_port = se->s_port;
    else
      sin.sin_port = htons(674); // hardcode the default port if all else fails
  }
  sin.sin_addr.s_addr = INADDR_ANY;

  // And create/bind the socket... */
  if ((sock = socket(PF_INET, SOCK_STREAM, 0)) < 0)
    throw UnixError("socket");
  {
    int opt = 1;
    if ((setsockopt(sock, SOL_SOCKET, SO_REUSEADDR,
                    (const char *)&opt, sizeof(opt))) < 0)
      throw UnixError("setsockopt");
  }

  if (bind(sock, (struct sockaddr *)&sin, sizeof(struct sockaddr_in))) {
    cerr << "Unable to listen on TCP port " << ntohs(sin.sin_port) << endl;
    throw UnixError("bind");
  }

  if (listen(sock, SOMAXCONN))
    throw UnixError("listen");

  return sock;
}
