/* conn.h -*- C++ -*- connection-level interface
 * Rob Earhart
 * $Id: conn.h,v 1.5 1998/03/09 00:58:00 rob Exp $
 */
/***********************************************************
        Copyright 1997 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of CMU not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.
******************************************************************/

#ifndef ACAP_CONN_H
#define ACAP_CONN_H

#include <fstream.h>
#include <netinet/in.h>
#include <hash_map.h>
#include <hash_set.h>
#include "lex.acap.h"
#include "ptr.h"
#include "refcount.h"
#include "mutex.h"
#include "void.h"

class Command;
class Language;
class Context;
class NotifyContext;

class Conn : public Refcounted {
public:
  Conn(int fd, struct sockaddr_in peer);
  ~Conn();
  void Loop() throw ();		// processes requests
  class ConnectionClosed {};
  class Logout {};

  ostream &Out() { return out; }
  Mutex &OutputMutex() { return output_mutex; }

  typedef hash_map<String, P<Context>,
                   hash<String>, equal_to<String> > contexts_t;
  contexts_t &Contexts() { return contexts; }
  Mutex &ContextsMutex() { return contexts_mutex; }

  int NextToken();
  void PutbackToken() { putback = true; }
  void Resync();
  ACAPLexer &Lexer() { return lexer; }

  // These're for authentication mechanisms, and anything else
  // in the future that's dynamically plugged into the server
  // and needs to access the parser in the middle of a command parse.
  String GetNextString();	// returns a string from stream
  void GetSpace();		// call when expecting a space
  void GetCRLF();		// call when expecting end-of-line
  void SetAuth(const String &s) {
    authentic = true;
    auth_id = s;
  }
  Mutex &AuthMutex() { return auth_mutex; }
  bool IsAuthentic() const { return authentic; }
  const struct sockaddr_in &GetPeerAddr() const { return peer; }
  const Language *CurrentLanguage() const { return language; }
  void SetLanguage(const Language *l) { language = l; }

  // This is a little esoteric... when a notifycontext is created that
  // wants to peer off a context, it creates an entry here, so that
  // when the named context is changed (created, recreated, freed,
  // whatever), the contexts peering off of it can update themselves.
  // All of these expect that the peer contexts mutex will be held.
  void AddPeer(const String &path, NotifyContext *context);
  void RemovePeer(const String &path, NotifyContext *context);
  void NotifyPeers(const String &path);
  Mutex &PeerContextsMutex() { return peer_contexts_mutex; }

private:
  int fd;
  struct sockaddr_in peer;
  Mutex output_mutex;
  ifstream in;
  ofstream out;
  ACAPLexer lexer;
  bool putback;
  int lasttoken;
  contexts_t contexts;
  Mutex contexts_mutex;
  Mutex auth_mutex;
  bool authentic;
  String auth_id;
  const Language *language;

  typedef hash_set<NotifyContext *, hash<void *>,
                   equal_to<void *> > peer_context_t;
  typedef hash_map<String, peer_context_t,
                   hash<String>, equal_to<String> > peer_contexts_t;
  peer_contexts_t peer_contexts;
  Mutex peer_contexts_mutex;
};

#endif // ACAP_CONN_H
