/* dataset.cc - dataset implementation
 * Rob Earhart
 * $Id: dataset.cc,v 1.9 1998/03/09 00:58:07 rob Exp $
 */
/***********************************************************
        Copyright 1998 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of CMU not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.
******************************************************************/

#include <deque.h>
#include <stack.h>
#include "dataset.h"
#include "dsgroup.h"
#include "entryref.h"
#include "msg.h"
#include "response.h"
#include "interlock.h"
#include "sort.h"
#include "notifycontext.h"

DatasetMap *DatasetMap::instance = NULL;

pthread_once_t DatasetMap::instance_init = PTHREAD_ONCE_INIT;

void DatasetMap::Initialize() {
  instance = new DatasetMap();
}

DatasetMap *DatasetMap::Instance() {
  if (! instance)
    pthread_once(&instance_init, &Initialize);
  return instance;
}

P<Dataset> DatasetMap::Lookup(const String &path, Trx *trx) {
  P<Dataset> result;
  DEBUG("Looking up dataset file " << path);
  {
    Grab g(datasets_mutex);
    pair<datasets_t::iterator, bool> p
      = datasets.insert(datasets_t::value_type(path, (Dataset *)NULL));
    if (! p.second)
      result = (*p.first).second;
    else {
      try {
	result = new Dataset(path);
	(*p.first).second = result;
      } catch (...) {
	datasets.erase(p.first);
	throw;
      }
    }
  }
  Grab g(result->InitMutex());
  result->Init(trx);
  return result;
}

P<Dataset> DatasetMap::CreateDataset(Entry entry, Trx *trx) {
  // Yeah, unsynchronized, but it *really* doesn't matter...
  // I could as easily use rand (in fact... hrm.)
  // XXX This is horribly inefficient...
  DEBUG("Creating new dataset");
  static unsigned long n = 0;
  char buf[PATH_MAX];
  int fd;
  while(1) {
    sprintf(buf, "ds%0l*x", sizeof(n) * 2, n++);
    int fd = open(buf, O_RDWR | O_CREAT | O_EXCL, 0600);
    if (fd >= 0) {
      close(fd);
      break;
    }
  }
  String path = String::Copy(buf);
  DEBUG("Picked path " << path);
  P<Dataset> result = new Dataset(path);
  result->Init(trx);
  DEBUG("Setting up subdataset's dataset.inherit attribute");
  Attribute ia = entry.Lookup("dataset.inherit", trx->TrxStamp());
  if (ia.Exist() && ia.IsString()) {
    // Use ia.StringValue() as the inheritance path
    trx->SetPS(result);
    Entry newent = result->FindEntry("", trx->TrxStamp());
    if (! newent.Exist())
      newent = result->InsertEntry("", NULL, trx);
    Attribute nia = newent.Lookup("dataset.inherit", trx->TrxStamp());
    if (nia.Exist()) nia.SetDelete(trx);
    nia = newent.CreateAttribute("dataset.inherit", trx);
    nia.SetSingleValue(trx, ia.StringValue());
    newent.SetModTime(trx->TrxStamp(), trx);
  } else {
    ia = entry.GetDataset()->FindEntry("", trx->TrxStamp()).Lookup("dataset.inherit", trx->TrxStamp());
    if (ia.Exist() && ia.IsString()) {
      // Build new inherit path
      String new_inherit = (String)ia.StringValue() + String("/") + (String)entry.Name();

      // Store it to the dataset...
      trx->SetPS(result);
      Entry newent = result->FindEntry("", trx->TrxStamp());
      if (! newent.Exist())
	newent = result->InsertEntry("", NULL, trx);
      ia = newent.Lookup("dataset.inherit", trx->TrxStamp());
      if (ia.Exist()) ia.SetDelete(trx);
      ia = newent.CreateAttribute("dataset.inherit", trx);
      ia.SetSingleValue(trx, new_inherit);
      newent.SetModTime(trx->TrxStamp(), trx);

      // Store it to the entry...
      trx->SetPS(entry.GetDataset());
      ia = entry.Lookup("dataset.inherit", trx->TrxStamp());
      if (ia.Exist()) ia.SetDelete(trx);
      ia = entry.CreateAttribute("dataset.inherit", trx);
      ia.SetSingleValue(trx, new_inherit);
    }
  }

  DEBUG("Setting subdataset attribute");
  trx->SetPS(entry.GetDataset());
  ia = entry.Lookup("", trx->TrxStamp());
  if (ia.Exist())
    ia.SetDelete(trx);
  ia = entry.CreateAttribute("", trx);
  ia.SetSingleValue(trx, path);
  ia = entry.Lookup("subdataset", trx->TrxStamp());
  if (ia.Exist())
    ia.SetDelete(trx);
  ia = entry.CreateAttribute("subdataset", trx);
  ia.SetSingleValue(trx, ".");
  entry.SetModTime(trx->TrxStamp(), trx);
  Grab g(datasets_mutex);
  datasets[path] = result;
  return result;
}

void DatasetMap::RemoveDataset(const String &path, Dataset *d) {
#ifdef NDEBUG
  datasets.erase(path);
#else  // NDEBUG
  datasets_t::iterator it = datasets.find(path);
  assert((*it).second == d);
  datasets.erase(it);
#endif // ! NDEBUG
}

Dataset::Dataset(const String &_path, int _fd) : PS(_path, _fd) {}

Dataset::~Dataset() {
  DatasetMap::Instance()->RemoveDataset(Path(), this);
}

void Dataset::Noref() {
  Grab g(DatasetMap::Instance()->DatasetsMutex());
  if (CheckNoRefs())
    delete this;
}

void Dataset::GetEntries(Dataset::dsentries_t &entries,
			 Timestamp::Stamp stamp,
			 DatasetGroup *group) {
    entrylist_t *ents = GetData()->entries;
    if (ents) {
      {
	P<PSBlob> blob = GetBlob(ents);
	group->AddBlob(blob);
      }
      for (unsigned lupe = 0;
	   lupe < ents->n_alloced;
	   lupe++) {
	entryval_t *ev = ents->buckets(lupe);
	if (ev->in_date(stamp) && ! (ev->name == ""))
	  entries.push_back(Entry(this, ev));
      }
    }
}

Entry Dataset::FindEntry(const String &name,
			 Timestamp::Stamp stamp) {
  DEBUG("Finding entry " << name << " in dataset " << this << " at " << stamp);
  entrylist_t *ents = GetData()->entries;
  if (ents)
    for (unsigned lupe = 0;
	 lupe < ents->n_alloced;
	 lupe++) {
      entryval_t *ev = ents->buckets(lupe);
      DEBUG("Checking entry " << lupe << " (" << DString(&ev->name)
	    << ": " << ev->createtime << ", " << ev->modtime << ", " << ev->deletetime << ')');
      if (ev->in_date(stamp)
	  && ev->name == name) {
	DEBUG("Match!");
	return Entry(this, ev);
      }
    }
  DEBUG("No matches");
  return Entry(this, NULL);
}

Entry Dataset::InsertEntry(const String &name,
			   Entry *oldentry,
			   Trx *trx) {
  DEBUG("Inserting entry for " << name);
  datasetobject_t *data = GetData();
  entrylist_t *ents = data->entries;

  if (! ents
      || ents->n_buckets == ents->n_alloced) {
    off_t newentsoffset;
    // Yeah, this makes insertion cost for large N quadratic.
    // XXX Figure out a better way to measure this...
    unsigned n_buckets = 0;
    if (ents) {
      for (unsigned lupe = 0;
	   lupe < ents->n_alloced;
	   lupe++) {
	entryval_t *ev = ents->buckets(lupe);
	  
	if (ev->in_date(trx->TrxStamp()))
	  n_buckets++;
      }
      n_buckets += 10;
    } else
      n_buckets = 20;
    entrylist_t *newents = new(this, n_buckets,
			       trx, &newentsoffset) entrylist_t(n_buckets, trx);
    if (ents) {
      P<PSBlob> eblob = ScheduleFree(ents, trx);
      for (unsigned lupe = 0;
	   lupe < ents->n_alloced;
	   lupe++) {
	entryval_t *ev = ents->buckets(lupe);
	bool is_old = oldentry && oldentry->IsSameData(ev);
	if (ev->in_date(trx->TrxStamp()) && ! is_old) {
	  entryval_t *nv = newents->alloc_bucket(trx);
	  nv->assign(*ev, this, trx);
	} else {
	  ev->schedule_free(this, eblob, ! is_old, trx);
	}
      }
    }
    data->set_entries(newents, trx, newentsoffset);
    ents = newents;
  }
  entryval_t *ev = ents->alloc_bucket(trx);
  ev->set_name(this, name, trx);
  Entry ent(this, ev);
  DEBUG("Created entry at " << ev);
  if (oldentry) {
    ent.AssignAttributes(*oldentry, trx);
    Attribute ia = ent.Lookup("entry", trx->TrxStamp());
    if (ia.Exist()) ia.SetDelete(trx);
    oldentry->SetDeleteTime(trx->TrxStamp(), trx);
  }
  Attribute nameattr = ent.CreateAttribute("entry", trx);
  nameattr.SetSingleValue(trx, name);
  ent.SetModTime(trx->TrxStamp(), trx);
  return ent;
}

void Dataset::GetDeleted(Dataset::names_t &deleted,
			 Dataset::names_t &defaulted,
			 Dataset::names_t &exist,
			 Timestamp::Stamp stamp,
			 Timestamp::Stamp synctime) {
  DEBUG("Getting deleted entries");
  entrylist_t *ents = GetData()->entries;
  if (ents) {
    for (unsigned lupe = 0;
	 lupe < ents->n_alloced;
	 lupe++) {
      Entry ev(this, ents->buckets(lupe));
      DEBUG("Checking status of entry " << lupe);
      if (synctime < ev.DeleteTime()) // we Just Don't Care about entries that were deleted
	if (ev.InDate(stamp))       // before the synctime...
	  if (ev.Visible()) {
	    DEBUG(ev.Name() << " appears to exist");
	    exist.insert(ev.Name());
	  } else {
	    DEBUG(ev.Name() << " appears to be deleted");
	    deleted.insert(ev.Name());
	  }
	else {
	  DEBUG(ev.Name() << " appears to be defaulted");
	  defaulted.insert(ev.Name());
	}
      else {
	DEBUG("The entry appears to be toast: synctime == " << synctime << "; deletetime == " << ev.DeleteTime());
      }
    }
  }
}

caddr_t Dataset::CreateRoot(Trx *trx, off_t *offset) {
  return (caddr_t) new(this, trx, offset) datasetobject_t(this, trx);
}

struct todo {
  todo(int _depth, const String &_relpath, P<Dataset> _dataset)
    : depth(_depth), relpath(_relpath), dataset(_dataset) {}
  int depth;
  String relpath;
  P<Dataset> dataset;
};

typedef hash_map<String, Ptr<EntryRef>, hash<String>, equal_to<String> > entry_names_t;

static void Augment(Dataset::dsentries_t &entries,
		    entry_names_t &entnames,
		    Timestamp::Stamp stamp) {
  for (Dataset::dsentries_t::iterator it = entries.begin();
       it != entries.end();
       ++it) {
    pair<entry_names_t::iterator, bool> p
      = entnames.insert(entry_names_t::value_type((*it).Name(), NULL));
    if (p.second) {
      // We actually inserted it
      try {
	(*p.first).second = new EntryRef(*it, stamp);
      } catch (...) {
	entnames.erase(p.first);
	throw;
      }
    } else
      (*p.first).second->AddInheritedEntry(*it);
  }
}

void Dataset::Search(Dataset::entries_t &result,
		     DatasetGroup *group,
		     NotifyContext *context,
		     const String &dspath,
		     Modifiers *modifiers,
		     Criteria *criteria,
		     Referer *referer,
		     Timestamp::Stamp stamp) {
  String prefixpath = dspath + "/";

  assert(result.empty());

  deque<Ptr<todo> > dq;		// depth-queue;
				// basically, a queue of the datasets
				// which need to be processed.

  // Okay.  This is the initializer for the depth recursion.
  Ptr<todo> first = new todo(modifiers ? modifiers->Depth() : 1,
			     "",
			     this);

  // And here's the recursion:
  for (dq.push_back(first); ! dq.empty(); dq.pop_front()) {
    todo *t = dq.front();
    // Now, we want to handle todo item t.
    // First, we have to get all of the entries for this dataset,
    // handling inheritance correctly.

    group->AddDataset(t->dataset);

    typedef hash_set<String, hash<String>, equal_to<String> > inherited_t;
    inherited_t inherited;
    entry_names_t names;
    entry_names_t::iterator it;

    {
      // Handle base dataset...
      dsentries_t entries;
      t->dataset->GetEntries(entries, stamp, group);
      Augment(entries, names, stamp);
      inherited.insert(dspath);
    }

    // And now go through all the inherited datasets.
    if (! modifiers || modifiers->Inherit()) {
      Attribute ia = t->dataset->FindEntry("", stamp).Lookup("dataset.inherit", stamp);
      while (ia.IsString()) {
	String inheritpath = ia.StringValue();
	if (! (inheritpath.size() && inherited.insert(inheritpath).second))
	  break;
	try {
	  P<Dataset> ds = TraversePath(inheritpath, stamp, NULL, context);
	  group->AddDataset(ds);
	  dsentries_t entries;
	  ds->GetEntries(entries, stamp, group); // get the entries
	  Augment(entries, names, stamp); // add them to our names table
	  ia = ds->FindEntry("", stamp).Lookup("dataset.inherit", stamp);
	} catch (Response::NoExist) {
	  break;
	}
      }
    }
    // At this point, our names map should be complete.
    // We walk through it, creating todo items... (since we need to add
    // all subdatasets before we can do criteria checking)
    if (t->depth == 0 || t->depth > 1) {
      bool added_path = false;
      for (it = names.begin();
	   it != names.end();
	   ++it) {
	Attribute subds = (*it).second->Lookup("subdataset");
	if (subds.Exist())
	  if (subds.IsString() && subds.StringValue() == ".") {
	    subds = (*it).second->Subdataset();
	    if (subds.Exist() && subds.IsString()) {
	      if (context) {
		if (! added_path) {
		  added_path = true;
		  context->AddPathDependancy(prefixpath + String(t->relpath.begin(), t->relpath.end() - 1));
		}
		context->AddDatasetDependancy(prefixpath + t->relpath + (*it).first);
	      }
	      P<Dataset> sub = DatasetMap::Instance()->Lookup(String::Copy(subds.StringValue()));
	      if (sub) {
		Ptr<todo> td = new todo(t->depth ? t->depth - 1 : 0,
					t->relpath + (*it).first + String("/"),
					sub);
		dq.push_back(td);
	      }
	    }
	  } else if (referer && (subds.IsString() || subds.IsMulti()) && subds.NValues() > 0) {
	    Attribute::values_t urls;
	    subds.MultiValue(urls);
	    (*referer)(prefixpath + t->relpath + (*it).first, urls);
	  }
      }
    }
    // And then we can copy the entries that pass our criteria check
    // into our final result vector.
    result.reserve(result.size() + names.size());
    for (it = names.begin();
	 it != names.end();
	 ++it)
      if ((*it).second->IsVisible() && ! (*criteria)((*it).second)) {
	// Save its path...
	if (modifiers && modifiers->GotDepth())
	  (*it).second->SetRelPath(prefixpath + t->relpath + (*it).first);
	else
	  (*it).second->SetRelPath((*it).first);
	// And save the entry itself.
	result.push_back((*it).second);
      }
    // And then continue on to the next item in the queue.
  }

  // Disgusted yet?  Now we sort the whole thing...
  // We do the weird casting here because we *don't* want
  // to be using Ptr<EntryRef>; introsort's partition step
  // copies out a value, which leaves a NULL behind, which
  // causes us to lose.
  if (modifiers && modifiers->GotSort())
    sort((vector<EntryRef *>::iterator)result.begin(),
	 (vector<EntryRef *>::iterator)result.end(),
	 Sort(modifiers->SortList()));

  // Assign entry numbers...
  for (unsigned ent_number = 0;
       ent_number < result.size();
       ent_number++)
    result[ent_number]->SetEntryNumber(ent_number + 1);

  // And we're done!  Wasn't that sick?
}

void Dataset::DeletedSince(Dataset::names_t &deleted,
			   DatasetGroup *dsgroup,
			   Timestamp::Stamp stamp,
			   Timestamp::Stamp synctime) {
  names_t::iterator it;
  assert(deleted.empty());

  // What we want to do is think about overlaying datasets.

  // We need to accumulate deleted, defaulted, and exist info for
  // each dataset down the inheritance tree.  Eventually, we need to
  // add everything that's been defaulted to the set of things that've
  // been deleted (since, for all intents and purposes, they're gone),
  // and then subtract off everything that actually exists.

  // The trick is that if something's been deleted in an inheriting dataset,
  // it doesn't matter if it exists in an inherited dataset - it's still
  // deleted, whereas if something's been defaulted in an inheriting dataset,
  // we still want to report it, unless it exists in an inherited dataset.

  // So we need to keep track of deleted, defaulted, and exist seperately.
  // We can add in defaulted and deleted each time, and add in exist
  // iff the name isn't in the original set of deleted items.  Then, at the
  // end, we can go back, add in the defaulted stuff, remove the stuff
  // that exists, and return that.

  // First, we get all our deleted, defaulted, and existing entry names...

  names_t defaulted, exist;
  dsgroup->AddDataset(this);
  GetDeleted(deleted, defaulted, exist, stamp, synctime);

  // Now we go through inheritance.
  Attribute ia = FindEntry("", stamp).Lookup("dataset.inherit", stamp);
  hash_set<String, hash<String>, equal_to<String> > inherited;
  while (ia.IsString()) {
    String ipath = ia.StringValue();
    if (! (ipath.size() && inherited.insert(ipath).second))
      break;
    try {
      names_t i_deleted, i_defaulted, i_exist;
      P<Dataset> ds = TraversePath(ipath, stamp);
      dsgroup->AddDataset(ds);
      ds->GetDeleted(i_deleted, i_defaulted, i_exist, stamp, synctime);
      ia = ds->FindEntry("", stamp).Lookup("dataset.inherit", stamp);

      // Now, we merge.

      // An entry goes from i_exist to exist if it's not already in deleted
      // (i.e. it exists in the inherited dataset, but it's been explicitly
      // deleted in the inheriting dataset).
      for (it = i_exist.begin();
	   it != i_exist.end();
	   ++it)
	if (deleted.find(*it) == deleted.end())
	  exist.insert(*it);

      // All entries in i_defaulted go into defaulted.
      for (it = i_defaulted.begin();
	   it != i_defaulted.end();
	   ++it)
	defaulted.insert(*it);

      // All entries in i_deleted go into deleted.
      for (it = i_deleted.begin();
	   it != i_deleted.end();
	   ++it)
	deleted.insert(*it);
    } catch (Response::NoExist) {
      break;
    }
  }

  // Then we add the ones that've been defaulted...
  for (it = defaulted.begin(); it != defaulted.end(); ++it)
    deleted.insert(*it);

  // And then we remove all the ones that exist (the original ones that exist,
  // plus anything in an inherited dataset that hasn't been deleted in an
  // inheriting dataset).
  for (it = exist.begin(); it != exist.end(); ++it)
    deleted.erase(*it);

  // Whew!
}

P<Dataset> Dataset::TraversePath(const String &basepath,
				 Timestamp::Stamp stamp,
				 Trx *trx,
				 NotifyContext *context) {
  DEBUG("Traversing path " << basepath);
  if (basepath.size() == 0) {
    if (context) context->AddDatasetDependancy(basepath);
    Interlock::Instance()->WaitForResource("", stamp, true);
    return DatasetMap::Instance()->Lookup("root", trx);
  }

  if (basepath[0] != '/')
    throw Response::NoExist(basepath, &Msg::IllegalPath);

  typedef stack<pair<String, StringTokenizer>, deque<pair<String, StringTokenizer> > > pathstack_t;
  pathstack_t pathstack;

  pathstack.push(pathstack_t::value_type(basepath, StringTokenizer(basepath, '/')));

  pathstack_t::value_type &base = pathstack.top();

  ++pathstack.top().second;	// skip the leading /
  DEBUG("Looking up root dataset");
  P<Dataset> cwd = DatasetMap::Instance()->Lookup("root", trx);

  while(! pathstack.empty()) {
    pathstack_t::value_type &cwp = pathstack.top();
    String &path = cwp.first;
    StringTokenizer &component = cwp.second;

    if (context) {
      context->AddDatasetDependancy(path);
      context->AddPathDependancy(path);
    }

    while (! component.Done()) {
      if ((*component).size() == 0)
	throw Response::NoExist(basepath, &Msg::IllegalPath);

      if (context && ! component.Last())
	context->AddPathDependancy(String::Copy(path.begin(),
						(*component).end() - path.begin()));

      DEBUG("cwd is " << cwd);
      DEBUG("Getting cwd resources");
      P<PSResources> rsrcs = cwd->GetResources();

      DEBUG("Looking up entry " << *component << " version " << stamp);
      Entry ent = cwd->FindEntry(*component, stamp);

      // Wait for the resource to become available...
      Interlock::Instance()->WaitForResource(String(path.begin(), (*component).end()),
					     stamp, ent.Exist());
      // In case it's changed...
      DEBUG("Re-looking up entry " << *component);
      ent = cwd->FindEntry(*component, stamp);
      if (! ent.Exist()) {
	DEBUG("The entry doesn't exist");
	if (trx) {
	  DEBUG("... but we can create it");
	  trx->SetPS(cwd);
	  ent = cwd->InsertEntry(*component, NULL, trx);
	} else {
	  DEBUG("... we can't create it; checking dataset.inherit");
	  // We can't create it, so we need to try virtual subdataset
	  // traversal.  Note that this same bit of code is duplicated
	  // below; I think people would kill me, though, if I wrapped
	  // it up in a macro, and it requires access to a few too
	  // many things to make it a function call, and I think it's
	  // clearer to have the code layed out here.
	  // ******** If it's changed, though, change it below, too... ********
	  Attribute ia = cwd->FindEntry("", stamp).Lookup("dataset.inherit", stamp);
	  if (ia.IsString()) {
	    // Okay - it's time to do virtual subdataset traversal...
	    DEBUG("dataset.inherit == " << ia.StringValue());
	    pathstack.push(pathstack_t::value_type(ia.StringValue(),
						   StringTokenizer(ia.StringValue(), '/')));
	    ++pathstack.top().second;
	    cwd = DatasetMap::Instance()->Lookup("root");
	    break;		// this will break us out to the packstack looping code
	  }
	  DEBUG("dataset.inherit is unuseable; entry does not exist; Punting");
	  // Otherwise, we didn't find the entry, and we're not inheriting,
	  // and we're not creating it, so we punt, using the last user-requested
	  // part of the path that we were unable to traverse.
	  throw Response::NoExist(String(base.first.begin(), (*base.second).end()),
				  &Msg::DatasetNoExist);
	}
      }
      DEBUG("Looking up subdataset attribute");
      {
	Attribute subds = ent.Lookup("subdataset", stamp);
	if (subds.Exist()
	    && (subds.IsString() || subds.IsMulti())
	    && ! (subds.IsString() && subds.StringValue() == ".")) {
	  // Whoops - referal time!
	  vector<String> urls;
	  Attribute::values_t durls;
	  subds.MultiValue(durls);
	  urls.reserve(durls.size());
	  for (Attribute::values_t::const_iterator it = durls.begin();
	       it != durls.end();
	       ++it) {
	    // This tacks the remaining part of the path onto the URL.
	    String url = *it;
	    if (url.size() && url[url.size() - 1] == '/')
	      // We don't need the '/' at the current component's end.
	      urls.push_back(url
			     + String((*base.second).end() + 1, base.first.end()));
	    else
	      // We *do* need the '/' at the current component's end.
	      urls.push_back(url
			     + String((*base.second).end(), base.first.end()));
	  }
	  throw Response::Refer(urls, &Msg::DatasetReferal);
	}
      }
      Attribute subdataset = ent.Subdataset(stamp);
      if (! subdataset.Exist() || subdataset.IsDefault()) {
	DEBUG("subdataset doesn't exist, or is default");
	if (trx) {
	  DEBUG("... but we can create it");
	  // First, blow away the default record
	  if (subdataset.Exist() && subdataset.IsDefault()) {
	    trx->SetPS(cwd);
	    subdataset.SetDelete(trx);
	    Attribute old = ent.Subdataset(stamp);
	    if (old.Exist())
	      old.SetDelete(trx); // XXX Decrement refcount on referenced dataset
	  }

	  // This will set the Subdataset magic attribute and the subdataset attribute.
	  cwd = DatasetMap::Instance()->CreateDataset(ent, trx);

	  // Since we just changed things, we let contexts know...
	  // Note that we have to do a path change for the path
	  // including the entry (in case anyone had searched through
	  // it), and a dataset change for the entry's dataset.
	  NotifyContext::NotifyPathChange(String(path.begin(), (*component).end()));
	  NotifyContext::NotifyDatasetChange(String(path.begin(), (*component).begin() - 1));

	  rsrcs = cwd->GetResources();
	} else {
	  DEBUG("... we can't create it; checking dataset.inherit");
	  // We can't create it, so we need to try virtual subdataset
	  // traversal.  Note that this same bit of code is duplicated
	  // above; I think people would kill me, though, if I wrapped
	  // it up in a macro, and it requires access to a few too
	  // many things to make it a function call, and I think it's
	  // clearer to have the code layed out here.
	  // ******** If it's changed, though, change it above, too... ********
	  Attribute ia = cwd->FindEntry("", stamp).Lookup("dataset.inherit", stamp);
	  if (ia.IsString()) {
	    // Okay - it's time to do virtual subdataset traversal...
	    DEBUG("dataset.inherit == " << ia.StringValue());
	    pathstack.push(pathstack_t::value_type(ia.StringValue(),
						   StringTokenizer(ia.StringValue(), '/')));
	    ++pathstack.top().second;
	    cwd = DatasetMap::Instance()->Lookup("root");
	    break;		// this will break us out to the packstack looping code
	  }
	  DEBUG("dataset.inherit is unuseable; subdataset does not exist; Punting");
	  // Otherwise, we didn't find the entry, and we're not inheriting,
	  // and we're not creating it, so we punt, using the last user-requested
	  // part of the path that we were unable to traverse.
	  throw Response::NoExist(String(base.first.begin(), (*base.second).end()),
				  &Msg::DatasetNoExist);
	}
      } else if (subdataset.IsNil()) {
	// We can't go on - there's explicitly no subdataset, so we punt.
	DEBUG("subdataset is explicitly NIL - aborting traversal");
	throw Response::NoExist(String(base.first.begin(), (*base.second).end()),
				&Msg::DatasetNoExist);
      } else {			// We got the subdataset, and can use it...
	assert(subdataset.IsString());
	DEBUG("Traversing to subdataset " << subdataset.StringValue());
	cwd = DatasetMap::Instance()->Lookup(String::Copy(subdataset.StringValue()));
      }
      do {
	DEBUG("Traversed " << *component);
	++component;		// go on to the next bit
      } while (! (*component).size() && ! component.Done());
    } // pathstack component loop
    pathstack.pop();		// go on to the next elt in the stack
  } // pathstack loop
  return cwd;			// nothing left to traverse, so...
}

void Dataset::SwizzleObject(caddr_t obj, unsigned char tag) {
  // I hate using a switch statement for this, but...
  switch(tag) {
  case empty_tag:
    //DEBUG("Ignoring empty tag at " << (void *)obj);
    break;
  case dataset_tag:
    //DEBUG("Swizzling dataset at " << (void *)obj);
    ((datasetobject_t *)obj)->Swizzle(this);
    break;
  case entrylist_tag:
    //DEBUG("Swizzling entrylist at " << (void *)obj);
    ((entrylist_t *)obj)->Swizzle(this);
    break;
  case string_tag:
    //DEBUG("Swizzling string at " << (void *)obj);
    ((string_t *)obj)->Swizzle(this);
    break;
  case attrlist_tag:
    //DEBUG("Swizzling attrlist at " << (void *)obj);
    ((attrlist_t *)obj)->Swizzle(this);
    break;
  case stringlist_tag:
    //DEBUG("Swizzling stringlist at " << (void *)obj);
    ((stringlist_t *)obj)->Swizzle(this);
    break;
  default:
    abort();			// uh oh - corrupt data?
  }
}

