/* entry.cc - common entry implementation
 * Rob Earhart
 * $Id: entry.cc,v 1.4 1998/03/09 00:58:17 rob Exp $
 */
/***********************************************************
        Copyright 1998 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of CMU not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.
******************************************************************/

#include "entry.h"
#include "dataset.h"

Attribute Entry::Lookup(const String &name, Timestamp::Stamp stamp) const {
  DEBUG("Looking up attr " << name);
  if (data && data->attrs)
    for (unsigned lupe = 0;
	 lupe < data->attrs->n_alloced;
	 lupe++) {
      attrval_t *av = data->attrs->buckets(lupe);
      if (av->in_date(stamp)
	  && av->name == name) {
	DEBUG("Matched attr " << lupe);
	return Attribute(dataset, av);
      }
    }
  DEBUG("No such attr");
  return Attribute(dataset, NULL);
}

// This probably could use some explaining.  :-)
// An entry may be inheriting or not inheriting
// (depending on whether the last full-entry store to it
// was a NIL or a DEFAULT; we default to DEFAULT).  On top of that,
// each individual attribute (even of a non-inheriting entry!)
// may be inheriting.  The catch: if an entry is non-inheriting,
// from that point onwards we can only inherit attributes that
// were defaulted IN THAT ENTRY.
// So: When we match, we have a filter.  If the filter is NULL,
// we can ignore it.  If, however, an entry is a non-inheriting
// entry, we insert into the filter all of the things that were
// default *in this entry*, that passed the old filter, and establish
// *that* as our new filter in subsequent matches.
void Entry::Match(attributes_t &result,
		  const String &pattern,
		  Timestamp::Stamp stamp,
		  Ptr<attr_filter_t> &prev_filter) const {
  Ptr<attr_filter_t> this_filter;
  if (! Inherit()) this_filter = new attr_filter_t;
  assert(pattern.size() && pattern[pattern.size() - 1] == '*');
  if (data && data->attrs)
    for (unsigned lupe = 0;
	 lupe < data->attrs->n_alloced;
	 lupe++) {
      Attribute av(dataset, data->attrs->buckets(lupe));
      DString name(av.Name());
      if (av.InDate(stamp)
	  && name.size()
	  && pattern.size() - 1 <= name.size()
	  && name != "modtime"
	  && (! prev_filter || prev_filter->find(name) != prev_filter->end())
	  && equal(pattern.begin(), pattern.end() - 1,
		   name.begin()))
	if (av.IsDefault()) {
	  if (! Inherit()) {
	    this_filter->insert(name);
	  }
	} else {
	  // Note that insert() here doesn't actually replace any attributes;
	  // it just inserts them if they're not already in the set, which takes
	  // care of our NIL masking (when we later filter out all the NILs)
	  // quite nicely.
	  result.insert(attributes_t::value_type(name, av));
	}
    }
  if (this_filter)		// implies !Inherit()
    prev_filter = this_filter;
  // The above is always correct: if prev_filter was NULL, we start filtering
  // on everything that was defaulted in this entry.  If prev_filter was not
  // NULL, then everything in this_filter must've been in prev_filter (or it
  // would've been masked out before being added to this_filter), so this_filter
  // is already the set difference between the prev_filter and the attributes
  // which were defaulted in this entry, which is exactly what we want to filter
  // upon in the future.  Whew!
}

Attribute Entry::CreateAttribute(const String &attribute,
				 Trx *trx) {
  DEBUG("Creating attribute for entry at " << data);
  attrlist_t *attrs = data->attrs;
  if (! attrs
      || attrs->n_buckets == attrs->n_alloced) {
    off_t newattrsoffset;
    // Yeah, this makes insertion cost for large N quadratic.
    // XXX Figure out a better way to measure this...
    unsigned n_buckets = 0;
    if (attrs) {
      for (unsigned lupe = 0;
	   lupe < attrs->n_alloced;
	   lupe++) {
	attrval_t *av = attrs->buckets(lupe);
	if (av->in_date(trx->TrxStamp()))
	  n_buckets++;
      }
      n_buckets += 10;
    } else
      n_buckets = 20;
    attrlist_t *newattrs
      = new(dataset, n_buckets,
	    trx, &newattrsoffset) attrlist_t(n_buckets, trx);
    DEBUG("Allocated new attrs at " << newattrs << " (offset " << (void *)newattrsoffset << ")");
    if (attrs) {
      DEBUG("Freeing old attrs at " << attrs);
      P<PSBlob> ablob = dataset->ScheduleFree(attrs, trx);
      for (unsigned lupe = 0;
	   lupe < attrs->n_alloced;
	   lupe++) {
	attrval_t *av = attrs->buckets(lupe);
	if (av->in_date(trx->TrxStamp())) {
	  attrval_t *nv = newattrs->alloc_bucket(trx);
	  nv->assign(*av, dataset, trx);
	} else {
	  av->schedule_free(dataset, ablob, trx);
	}
      }
    }
    data->set_attrs(newattrs, trx, newattrsoffset);
    attrs = newattrs;
  }
  attrval_t *av = attrs->alloc_bucket(trx);
  av->set_name(dataset, attribute, trx);
  return Attribute(dataset, av);
}
