/* entryref.cc - entryref implementation
 * Rob Earhart
 * $Id: entryref.cc,v 1.2 1998/03/09 00:58:19 rob Exp $
 */
/***********************************************************
        Copyright 1998 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of CMU not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.
******************************************************************/

#include "entryref.h"
#include "clientstr.h"

static void ProcessAttr(ostream &out,
			Attribute attr,
			const String &default_name,
			Modifiers::Metadata::metadata_t &metadata) {
  bool initspace;
  if (metadata.size() > 1) {
    initspace = false;
    out << " (";
  } else
    initspace = true;
  for (Modifiers::Metadata::metadata_t::iterator mit = metadata.begin();
       mit != metadata.end();
       ++mit) {
    if (initspace)
      out << ' ';
    else
      initspace = true;
    switch(*mit) {
    case Token::A_ATTRIBUTE:
      if (attr.Exist())
	out << ClientStr(attr.Name());
      else
	out << ClientStr(default_name);
      break;
    case Token::A_MYRIGHTS:
      out << "NIL";
      break;
    case Token::A_SIZE:
      if (! attr.IsValued() || attr.IsMulti() && attr.NValues() == 0)
	out << "0";
      else if (attr.IsString())
	out << attr.StringValue().size();
      else {
	out << '(';
	Attribute::values_t values;
	attr.MultiValue(values);
	for (Attribute::values_t::const_iterator it = values.begin();
	     it != values.end();
	     ++it) {
	  if (it != values.begin())
	    out << ' ';
	  out << (*it).size();
	}
	out << ')';
      }
      break;
    case Token::A_COUNT:
      out << (attr.IsValued() ? attr.NValues() : 0);
      break;
    case Token::A_VALUE:
      if (! attr.IsValued())
	out << "NIL";
      else if (attr.IsString())
	out << ClientStr(attr.StringValue());
      else {
	out << '(';
	Attribute::values_t values;
	attr.MultiValue(values);
	for (Attribute::values_t::const_iterator it = values.begin();
	     it != values.end();
	     ++it) {
	  if (it != values.begin())
	    out << ' ';
	  out << ClientStr(*it);
	}
	out << ')';
      }
      break;
    case Token::A_ACL:
      if (! attr.IsValued())
	out << "NIL";
      else {
	out << '(';
	Attribute::acls_t acls;
	attr.ACLValue(acls);
	for (Attribute::acls_t::const_iterator it = acls.begin();
	     it != acls.end();
	     ++it) {
	  if (it != acls.begin())
	    out << ' ';
	  out << ClientStr(*it);
	}
	out << ')';
      }
      break;
    case Token::A_CONTENTTYPE:
      if (! attr.IsValued() || ! attr.HaveContentType())
	out << "NIL";
      else
	out << ClientStr(attr.ContentType());
      break;
    }
  }
  if (metadata.size() > 1)
    out << ')';
}

void EntryRef::Write(ostream &stream,
		     Modifiers::returnlist_t::const_iterator it,
		     Modifiers::returnlist_t::const_iterator finish) const {
  for (; it != finish; ++it) {
    const String &name = (*it).first;
    if (name[name.size() - 1] != '*') {
      Attribute attr = Lookup(name);
      ProcessAttr(stream, attr, name, (*it).second->GetMetadata());
    } else {
      attributes_t attrs;
      Match(attrs, name);
      for (attributes_t::const_iterator ait = attrs.begin();
	   ait != attrs.end();
	   ++ait)
	ProcessAttr(stream, (*ait).second, (*ait).first, (*it).second->GetMetadata());
    }
  }
}

void EntryRef::Match(EntryRef::attributes_t &result,
		     const String &pattern) const {
  Ptr<Entry::attr_filter_t> filter = NULL;
  for (ents_t::const_iterator it = ents.begin();
       it != ents.end() && ! (filter && filter->empty());
       ++it)
    (*it).Match(result, pattern, stamp, filter);
  // We may have built up some NILs in the process (in order for entry attributes
  // to mask inherited entry attributes), so let's eliminate them:
  {
    // ? why in the WORLD doesn't the STL have a "remove_if" operation
    // defined as part of the container concept?  It's a fairly basic
    // operation, but one that requires knowledge of the container's
    // structure; you can't really do it well with a simple pair
    // iterators, so it really should be a container
    // method... grumble.
    attributes_t::iterator it = result.begin();
    while(it != result.end())
      if ((*it).second.IsNil()) {
	attributes_t::iterator del = it;
	++it;			// we advance before we erase, because erasing will
	result.erase(del);	// invalidate the current iterator.
      } else
	++it;
  }

  // Check modtime...
  if (pattern.size()
      && pattern.size() < 8
      && equal(pattern.begin(), pattern.end() - 1, "modtime")) {
    Attribute modtime = Lookup("modtime");
    result.insert(attributes_t::value_type(modtime.Name(), modtime));
  }
}

Timestamp::Stamp EntryRef::Stamp() const {
  return stamp;
}

Timestamp::Stamp EntryRef::ModTime() const {
  Timestamp::Stamp result;
  for (ents_t::const_iterator it = ents.begin();
       it != ents.end();
       ++it) {
    Timestamp::Stamp s = (*it).ModTime();
    if (result < s)
      result = s;
    if (! (*it).Inherit() && ! (*it).HasADefaultAttribute(stamp))
      break;			// Theoretically, we only care about modtimes
				// for attributes which might be inherited;
				// however, computing that would be MUCH more
				// annoying.
  }
  return result;
}

Attribute EntryRef::Lookup(const String &name) const {
  bool ismodtime(name == "modtime");
  Attribute result(NULL, NULL);
  Timestamp::Stamp latest;
  for (ents_t::const_iterator it = ents.begin();
       it != ents.end();
       ++it) {
    if (ismodtime) {
      Attribute checkme = (*it).Lookup(name, stamp);
      if (checkme.Exist() && checkme.IsString() && latest < checkme.CreateTime()) {
	latest = checkme.CreateTime();
	result = checkme;
      }
      if (! (*it).Inherit() && ! (*it).HasADefaultAttribute(stamp)) break;
    } else {
      result = (*it).Lookup(name, stamp);
      if (! result.Exist())	// if the entry doesn't exist...
	if ((*it).Inherit())	// and we're inheriting...
	  continue;		// keep trying...
	else			// otherwise...
	  break;		// stop now
				// so here, the entry exists...
      if (! result.IsDefault())	// so if it's not defaulted...
	return result;		// return it...
				// otherwise, keep trying (if it's defaulted,
				// we don't care whether or not the actual
				// entry's inheriting).
    }
  }
  return result;
}

bool EntryRef::IsVisible() const {
  for (ents_t::const_iterator it = ents.begin();
       it != ents.end();
       ++it) {
    if ((*it).Visible()) return true;
    if (! (*it).Inherit() /* and it's not visible */) return false;
  }
  return false;			// everything's defaulted?!?
}
