/* lex.acap.l - a -*- C++ -*- flex acap protocol parser
 * Rob Earhart
 * $Id: lex.acap.l,v 1.4 1998/03/09 00:58:31 rob Exp $
 */
/***********************************************************
        Copyright 1997 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of CMU not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.
******************************************************************/

/* This is *very* flex-specific... */

%option 8bit case-insensitive always-interactive nostdinit noyywrap nodefault
%option read c++ prefix="acap" yyclass="ACAPLexer" debug

%{
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#define ACAP_HAVE_ACAP_LEXER
#include "lex.acap.h"
#include "token.h"
#include "error.h"
#include "conn.h"
#include "grab.h"
#include "debug.h"

#define YY_BREAK

#define TOO_LONG_MESSAGE "You're not allowed to send a literal that long (according to the protocol), and attempting to do so in a non-synchronizing manner basically means we lose synchronization; we don't have a prayer of talking to each other anymore, so I'm dropping the connection.  Sorry..."

%}

CRLF    (\r\n)
SPACE   ([ ])
CTL     ([\x00-\x1f\x7f])
SAFE_CHAR ([\x01-\x09\x0b\x0c\x0e-\x21\x23-\x5b\x5d-\x7f])
QUOTED_SPECIALS ([\"\\])
TAG_CHAR ([^\(\)\{\%\* \x00-\x1f\x7f-\xff\"\\\+])
ATOM_CHAR ([^\(\)\{\% \x00-\x1f\x7f-\xff\"\\])

%x T
%x RECOVERY

%%

<INITIAL,T,RECOVERY><<EOF>> {
  throw Conn::ConnectionClosed();
}

<T>{TAG_CHAR}{1,32}/{SPACE} {
  // We only match tags in this state.
  BEGIN(0);
  return Token::TAG;
}
<T>.|\n {
  // All non-tags in the tag state are garbage.
  yyless(0);
  BEGIN(RECOVERY);
  break;
}

<RECOVERY>"{"[[:digit:]]+"+}"{CRLF} {
  unsigned long count;
  errno = 0;
  count = strtoul(yytext+1, NULL, 10);
  if (errno)
    // The client is being braindamaged - there's really no good way
    // to resynchronize the stream (well, I guess we could read over
    // ULONG_MAX characters, but...)
    throw Error("lexer", TOO_LONG_MESSAGE);
  while(count--) {
    if (yyinput() == EOF)
      break;
  }
  break;
}
<RECOVERY>{CRLF} {
  return Token::GARBAGE;
}
<RECOVERY>.|\n	{
  break; /* Just keep eating characters... */
}

{SPACE}		{ return Token::SPACE; }
{CRLF}		{ return Token::CRLF; }
addto		{ return Token::A_ADDTO; }
all		{ return Token::A_ALL; }
and		{ return Token::A_AND; }
authenticate	{ return Token::A_AUTHENTICATE; }
change		{ return Token::A_CHANGE; }
compare		{ return Token::A_COMPARE; }
comparestrict	{ return Token::A_COMPARESTRICT; }
default		{ return Token::A_DEFAULT; }
delete		{ return Token::A_DELETE; }
deleteacl	{ return Token::A_DELETEACL; }
deletedsince	{ return Token::A_DELETEDSINCE; }
depth		{ return Token::A_DEPTH; }
enumerate	{ return Token::A_ENUMERATE; }
equal		{ return Token::A_EQUAL; }
freecontext	{ return Token::A_FREECONTEXT; }
getacl		{ return Token::A_GETACL; }
hardlimit	{ return Token::A_HARDLIMIT; }
lang		{ return Token::A_LANG; }
limit		{ return Token::A_LIMIT; }
listrights	{ return Token::A_LISTRIGHTS; }
lock		{ return Token::A_LOCK; }
login		{ return Token::A_LOGIN; }
logout		{ return Token::A_LOGOUT; }
makecontext	{ return Token::A_MAKECONTEXT; }
modtime		{ return Token::A_MODTIME; }
myrights	{ return Token::A_MYRIGHTS; }
nil		{ return Token::A_NIL; }
nocreate	{ return Token::A_NOCREATE; }
noinherit	{ return Token::A_NOINHERIT; }
noop		{ return Token::A_NOOP; }
not		{ return Token::A_NOT; }
notify		{ return Token::A_NOTIFY; }
notifycontext	{ return Token::A_NOTIFYCONTEXT; }
or		{ return Token::A_OR; }
prefix		{ return Token::A_PREFIX; }
quit		{ return Token::A_QUIT; }
range		{ return Token::A_RANGE; }
removefrom	{ return Token::A_REMOVEFROM; }
return		{ return Token::A_RETURN; }
search		{ return Token::A_SEARCH; }
setacl		{ return Token::A_SETACL; }
sort		{ return Token::A_SORT; }
store		{ return Token::A_STORE; }
substring	{ return Token::A_SUBSTRING; }
unchangedsince	{ return Token::A_UNCHANGEDSINCE; }
unlock		{ return Token::A_UNLOCK; }
updatecontext	{ return Token::A_UPDATECONTEXT; }
"*" 		{ return Token::STAR; }
[[:alpha:]]{ATOM_CHAR}{0,20} {
  // 20 was chosen here 'cause it's relatively small, but longer than
  // the longest atom, so this'll match something like
  // "authenticatelogout", which we do NOT want to have translated
  // into the two seperate tokens.  Presumably, once we return a
  // Token::GARBAGE, everything else, even if it matches something,
  // will be thrown away.
  // We don't need a yyless here, 'cause there's no chance that
  // we've got a CRLF (or even a CR) in yytext.
  // Of course, if it's truly long, we'll fall out into the .|\n rule,
  // which will do the same thing.
  BEGIN(RECOVERY);
  break;
}

[[:digit:]]+ {
  return Token::NUMBER;
}

\"({SAFE_CHAR}|(\\{QUOTED_SPECIALS}))*\" {
  return Token::QSTRING;
}

\{[[:digit:]]+\+?\}{CRLF} {
  errno = 0;
  literal_length = strtoul(yytext+1, NULL, 10);
  if (errno)
    throw Error("lexer", TOO_LONG_MESSAGE);

  nonsync_literal = (yytext[yyleng-4] == '+');

  return Token::LITERAL;
}

"("		{ return Token::LPAREN; }
")"		{ return Token::RPAREN; }

.|\n		{ yyless(0); BEGIN(RECOVERY); break; }

%%

void ACAPLexer::ExpectingTag() {
  BEGIN(T);
}

void ACAPLexer::Resync() {
  BEGIN(RECOVERY);
  if (literal_length && nonsync_literal) {
    // We need to chow down on this literal
    while(literal_length--)
      if (yyinput() == EOF)
	break;
  }
  yylex();			// this will return a Token::GARBAGE,
				// after which we're resynced.
}

String ACAPLexer::Literal() {
  if (! nonsync_literal) {
    Grab g(conn->OutputMutex());
    conn->Out() << "+ \"Ready for data\"\r\n" << flush;
  }

  String literal(literal_length);

  for (String::value_type *it = literal.octets();
       literal_length;
       literal_length--, it++) {
    int c = yyinput();
    if (c == EOF) throw Conn::Logout();
    *it = c;
  }

  return literal;
}

String ACAPLexer::Quoted() {
  char *start = yytext + 1;	// point at first data char
  char *src = start;
  char *dest = start;
  char *end = yytext + yyleng - 1; // point at close quote
  while(src < end) {
    if (*src == '\\')
      src++;
    *dest++ = *src++;
  }
  return String::Copy(start, dest - start);
}
