/* search.cc - search command implementation
 * Rob Earhart
 * $Id: search.cc,v 1.7 1998/03/09 00:58:45 rob Exp $
 */
/***********************************************************
        Copyright 1997 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of CMU not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.
******************************************************************/

#include <strstream.h>
#include "response.h"
#include "search.h"
#include "interlock.h"
#include "grab.h"
#include "dataset.h"
#include "conn.h"
#include "notifycontext.h"
#include "staticcontext.h"
#include "clientstr.h"
#include "dsgroup.h"
#include "entryref.h"

static Ptr<Criteria> subcrit = new CritEQUAL("subdataset", "i;octet", NULL);
Ptr<Criteria> Search::subdataset_criteria = new CritNOT(subcrit);

Search::Search(const String &_tag, Conn *_conn, const String &_path)
  : Command(_tag, _conn), path(_path), ncontext(NULL) {
    modifiers = new Modifiers();
    if (path.size() < 1)
      throw Response::BAD("Zero-length dataset/context path specified");
    if (path[0] != '/')
      Context::CheckLegalContextName(path);
    else {
      if (path == "/") path = "";
      if (path.size() && path[path.size() - 1] == '/')
	throw Response::BAD("Search paths shouldn't end in '/'s");
    }
}

void Search::Setup() {
  ts = Interlock::Instance()->CreateStamp();
  if (! Context::IsDatasetPath(path)) {
    // If we're using a context, we need to look it up now,
    // since subsequent commands (including this one!)
    // may change the name->context mapping.  (JKMN)
    Grab g(Client()->ContextsMutex());
    Conn::contexts_t::iterator it = Client()->Contexts().find(path);
    if (it == Client()->Contexts().end())
      throw Response::BAD(&Msg::NoSuchContext);
    searchable = (*it).second;
    is_enum_context = (*it).second->Enumerated();
  } else {
    // We're searching a dataset; it's definitely *not* an enumerated context!  :-)
    is_enum_context = false;
  }

  // And now we can set up our context...
  if (modifiers->GotMakeContext()) {
    if (modifiers->Notify()) {
      ncontext = new NotifyContext(modifiers->Context(),
				   Client(),
				   ts->GetStamp(),
				   path,
				   modifiers->Enumerate());
      context = ncontext;
      if (! modifiers->GotSort()) {
	// In order to do notification, we're going to need to sort
	// the data anyway at some point; if we don't impose some
	// ordering upon the data, we'll just have to sort it later as we
	// make it into sets, and we won't be able to have enumeration
	// (since the order of the data will be undefined, and we can't really
	// mix straight set theory with context enumeration offset management without
	// some very annoying and slow code).
	// So we might as well just define a sort now and get it over with -
	// we're allowed to do this, and it'll speed things up later, as well
	// as simplify the code.
	modifiers->SettingSortList();
	modifiers->AddSort("entry", "+i;octet");
	modifiers->DoneWithSortList();
      }
    } else
      context = new StaticContext(Client(),
				  ts->GetStamp(),
				  modifiers->Enumerate());
    {
      Grab g(Client()->ContextsMutex());
      Client()->Contexts()[modifiers->Context()] = context;
    }
    context_grab = new Grab(context->StateMutex());
    {
      Grab g(Client()->PeerContextsMutex());
      Client()->NotifyPeers(modifiers->Context());
    }
  }
}

class SearchReferer : public Searchable::Referer {
public:
  SearchReferer(Conn *_client, const String &_tag) : client(_client), tag(_tag) {}
  void operator()(const String &path,
		  const vector<DString> &urls) const {
    Grab g(client->OutputMutex());
    ostream &out = client->Out();
    out << tag << " Refer " << ClientStr(path);
    for (vector<DString>::const_iterator it = urls.begin();
	 it != urls.end();
	 ++it)
      out << ' ' << ClientStr(*it);
    out << "\r\n";
  }
private:
  Conn *client;
  String tag;
};

void Search::Eval() {
  if (! searchable) {
    // We have to do this as part of the downcast
    P<Dataset> dataset = Dataset::TraversePath(path, ts->GetStamp(), NULL, ncontext);
    searchable = dataset;
  }
  Ptr<DatasetGroup> dsgroup = new DatasetGroup();

  Searchable::entries_t entries;
  SearchReferer referer(Client(), Tag());
  searchable->Search(entries, dsgroup, ncontext, path, modifiers, criteria, &referer, ts->GetStamp());

  for (Searchable::entries_t::const_iterator it = entries.begin();
       it != entries.end();
       ++it) {
    if (modifiers->GotReturn()) {
      // XXX We really should buffer independantly, instead of
      // grabbing the stream while we read from disk, etc.
      // Solaris's ostrstream implementation (which I was using
      // for buffering) appears to be broken, though, so instead,
      // we'll settle for grabbing/releasing every cycle, giving
      // other threads a chance to send stuff.  This should work
      // fairly well, even though it's a bit suboptimal...
      Grab g(Client()->OutputMutex());
      ostream &out = Client()->Out();
      out << Tag() << " Entry " << ClientStr((*it)->RelPath());
      (*it)->Write(out,
		   modifiers->ReturnList().begin(),
		   modifiers->ReturnList().end());
      out << "\r\n";
    }
  }
  Grab g(Client()->OutputMutex());
  Client()->Out() << Tag() << " Modtime \"" << ts->LongStr() << "\"\r\n" << flush;

  if (context)
    context->DoneInitializing(entries, dsgroup, criteria, modifiers);
}

Search::~Search() {
  if (context && context->State() == Context::bad) {
    Grab g(Client()->ContextsMutex());
    if (Client()->Contexts()[modifiers->Context()] == context)
      Client()->Contexts().erase(modifiers->Context());
  }
}

Msg::Generator Search::Name() const {
  return &Msg::Search;
}
