/* store.cc - store command implementation
 * Rob Earhart
 * $Id: store.cc,v 1.7 1998/03/09 00:58:52 rob Exp $
 */
/***********************************************************
        Copyright 1997 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of CMU not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.
******************************************************************/

#include <algo.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <fcntl.h>
#include <strstream.h>
#include "response.h"
#include "store.h"
#include "conn.h"
#include "error.h"
#include "grab.h"
#include "interlock.h"
#include "unixerror.h"
#include "util.h"
#include "dataset.h"
#include "notifycontext.h"

void Store::Setup() {
  // We have to block all other stores from the time when
  // we get the timestamp to the time when we are in all
  // of the queues.  It actually doesn't suck all that badly;
  // we drop the mutex before we ever do anything that could
  // block us (like talking to the disk).
  Grab grab(Interlock::Instance()->InterlockMutex());
  version = new Interlock::Version();
  DEBUG("We are store version " << version->GetTimestamp()->LongStr());
  
  // Here, we add ourselves to every entry
  // that we're interested in (while we've got the interlock mutex).
  for (entries_t::iterator it = entries.begin();
       it != entries.end();
       ++it) {
    // XXX Actually, we only really have to lock (*it).second->Name(),
    // and we could achieve more parallism that way.  We'd have to
    // check each entry as we looked it up, and we'd need to change
    // FSDataset::AppendEntry to be able to handle simultaneous writers
    // to different entries in the same dataset.
    const String &name = (*it).second->DatasetName();
    version->AddResource(name, false);
    if (! (*it).second->NoCreate()) {
      // We need to record that we'll be trying to implicitly create
      // all of the intervening datasets.
      for (String::const_iterator it = name.begin();
	   it != name.end();
	   ++it)
	if (*it == '/')
	  version->AddResource(String(name.begin(), it), true);
    }
  }
  // popping out here drops our lock on the Interlock mutex.
}

void Store::AddEntry(Ptr<StoreEntry> &ent) {
  pair<entries_t::iterator, bool> b
    = entries.insert(entries_t::value_type(ent->Name(), ent));
  if (! b.second)
    throw Response::BAD("Entry listed multiple times");
}

Msg::Generator Store::Name() const { return &Msg::Store; }

static void ApplyAttribute(Entry entry,
			   Trx *trx,
			   const String &name,
			   const StoreEntry::Attribute *a) {
  {
    Attribute oldattr = entry.Lookup(name, trx->TrxStamp());
    if (oldattr.Exist())
      oldattr.SetDelete(trx);
  }
  // XXX Handle other metadata
  if (! a || a->GetValue()) {
    switch (a ? a->GetValue()->GetValueType() : StoreEntry::Value::NIL) {
    case StoreEntry::Value::NIL:
      if (entry.Inherit()) {
	Attribute newattr
	  = entry.CreateAttribute(name, trx);
	newattr.SetNilValue(trx);
      }
      break;
    case StoreEntry::Value::DEFAULT:
      if (! entry.Inherit()) {
	Attribute newattr
	  = entry.CreateAttribute(name, trx);
	newattr.SetDefaultValue(trx);
      }
      break;
    case StoreEntry::Value::STRING:
      {
	Attribute newattr
	  = entry.CreateAttribute(name, trx);
	newattr.SetSingleValue(trx, a->GetValue()->GetValues()[0]);
      }
    break;
    case StoreEntry::Value::STRINGLIST:
      {
	Attribute newattr
	  = entry.CreateAttribute(name, trx);
	newattr.SetMultiValue(trx, a->GetValue()->GetValues());
      }
    break;
    default:
      throw InternalError("Bad store-metadata type discovered during store");
    }
  }
}

void Store::Eval() {
  DEBUG("Evaluating store command");

  String modtime(version->GetTimestamp()->LongStr());

  Ptr<Trx> trx = new Trx(version->GetTimestamp());

  entries_t::iterator it;
  for (it = entries.begin();
       it != entries.end();
       ++it) {
    StoreEntry *e = (*it).second;
    DEBUG("Traversing entry " << e->Name());
    P<Dataset> dataset
      = Dataset::TraversePath(e->DatasetName(), version->GetTimestamp()->GetStamp(),
			      e->GetNoCreate() ? NULL : trx);
    DEBUG("Traversed path to " << e->DatasetName());
    dsgroup.AddDataset(dataset);
    DEBUG("Finding entry " << e->EntryName());
    Entry entry = dataset->FindEntry(e->EntryName(), version->GetTimestamp()->GetStamp());

    // check stuff
    e->SetDSEntry(dataset, entry);
  }

  DEBUG("Paths traversed - updating...");

  for (it = entries.begin();
       it != entries.end();
       ++it) {
    // write stuff
    StoreEntry *e = (*it).second;
    Dataset *dataset = e->GetDataset();
    Entry entry = e->GetEntry();

    DEBUG("Updating entry " << e->Name());

    NotifyContext::NotifyDatasetChange(e->DatasetName());
    if (e->SubdatasetAttribute() && e->SubdatasetAttribute()->GetValue()
	|| e->EntryAttribute() && e->EntryAttribute()->GetValue()
	|| e->InheritAttribute() && e->InheritAttribute()->GetValue())
      NotifyContext::NotifyPathChange(e->Name());

    trx->SetPS(dataset);

    if (e->EntryAttribute() && e->EntryAttribute()->GetValue()) {
      StoreEntry::Value *value = e->EntryAttribute()->GetValue();
      switch (value->GetValueType()) {
      case StoreEntry::Value::NIL:
	// Storing a NIL replaces the entry with an empty, non-inheriting
	// placeholder.
	{
	  if (entry.Exist()) entry.SetDeleteTime(trx->TrxStamp(), trx);
	  Entry newentry = dataset->InsertEntry(e->EntryName(), NULL, trx);
	  newentry.SetInherit(false, trx);
	  continue;		// just go on to the next entry - nothing to see here...
	}
      case StoreEntry::Value::DEFAULT:
	{
	  // Storing a DEFAULT removes the entry - it no longer exists.
	  // If it is stored to later, it will become an inheriting
	  // entry, as though it had never existed.
	  if (entry.Exist()) entry.SetDeleteTime(trx->TrxStamp(), trx);
	  continue;		// just go on - nothing more to do...
	}
      case StoreEntry::Value::STRING:
	{
	  // Changing a name removes the original entry and the
	  // destination entry, and creates a new entry with a copy of
	  // all attributes which won't be set in this trx.
	  Entry oldentry = dataset->FindEntry(value->GetValues()[0], trx->TrxStamp());
	  if (oldentry.Exist())
	    oldentry.SetDeleteTime(trx->TrxStamp(), trx);

	  Entry newentry = dataset->InsertEntry(value->GetValues()[0],
						entry.Exist() ? &entry : NULL,
						trx);
	  if (entry.Exist() && ! entry.Inherit())
	    newentry.SetInherit(false, trx);
	  entry = newentry;
	  NotifyContext::NotifyPathChange(e->DatasetName() + String("/") + value->GetValues()[0]);
	  break;			// go on and add attributes
	}
      case StoreEntry::Value::STRINGLIST:
	throw InternalError("Rename to multivalued string detected during store");
      default:
	throw InternalError("Bad store-metadata type discovered during store");
      }
    } else if (! entry.Exist()) {
      // Otherwise, we have to create it...
      entry = dataset->InsertEntry(e->EntryName(), NULL, trx);
    }

    if (e->SubdatasetAttribute() && e->SubdatasetAttribute()->GetValue()) {
      const StoreEntry::Attribute *attr = e->SubdatasetAttribute();
      StoreEntry::Value *value = attr->GetValue();
      switch (value->GetValueType()) {
      case StoreEntry::Value::NIL:
      case StoreEntry::Value::DEFAULT:
	ApplyAttribute(entry, trx, "subdataset", attr);
	ApplyAttribute(entry, trx, "", NULL);
	// Decrement refcount of associated dataset
	break;
      case StoreEntry::Value::STRING:
	if (value->GetValues()[0] == ".")
	  DatasetMap::Instance()->CreateDataset(entry, trx);
	else {
	  ApplyAttribute(entry, trx, "subdataset", attr);
	  ApplyAttribute(entry, trx, "", NULL);
	  // Decrement refcount of associated dataset
	}
	break;
      case StoreEntry::Value::STRINGLIST:
	ApplyAttribute(entry, trx, "subdataset", attr);
	ApplyAttribute(entry, trx, "", NULL);
	// Decrement refcount of associated dataset
	break;
      default:
	throw InternalError("Bad store-metadata type discovered during store");
      }
    }

    if (e->InheritAttribute()
	&& e->InheritAttribute()->GetValue()) {
      // Okay; the trick here is that if the entry has a subdataset,
      // we need to store the inherit attribute into the subdataset.
      // Note that since InheritAttribute is also in the attributes set,
      // we don't have to directly apply it to this dataset.
      DEBUG("Processing inheritance...");
      DEBUG("Looking up entry's magic attribute...");
      Attribute attr = entry.Lookup("", trx->TrxStamp());
      if (attr.Exist() && attr.IsString()) {
	DEBUG("Found it; traversing...");
	P<Dataset> newds = DatasetMap::Instance()->Lookup(String::Copy(attr.StringValue()));
	P<PSResources> rsrcs = newds->GetResources();
	trx->SetPS(newds);
	DEBUG("Finding magic entry in subdataset...");
	Entry newent = newds->FindEntry("", trx->TrxStamp());
	if (! newent.Exist())
	  newent = newds->InsertEntry("", NULL, trx);
	DEBUG("Applying dataset.inherit change");
	ApplyAttribute(newent, trx, "dataset.inherit", e->InheritAttribute());
	trx->SetPS(dataset);
	NotifyContext::NotifyDatasetChange(e->Name());
      }
    }

    for (StoreEntry::attributes_t::const_iterator attr
	   = e->Attributes().begin();
	 attr != e->Attributes().end();
	 ++attr) {
      const String &name = (*attr).first;
      StoreEntry::Attribute *a = (*attr).second;
      ApplyAttribute(entry, trx, (*attr).first, (*attr).second);
    }

    if (entry.ModTime() != trx->TrxStamp())
      entry.SetModTime(trx->TrxStamp(), trx);
  }

  trx->Commit();
  DEBUG("Store command complete");
}
