/* str.h -*- C++ -*- string implementation
 * Rob Earhart
 * $Id: str.h,v 1.6 1998/02/18 21:20:02 rob Exp $
 */
/***********************************************************
        Copyright 1997 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of CMU not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.
******************************************************************/

#ifndef ACAP_STR_H
#define ACAP_STR_H

// We need some sort of length-string interface;
// vanilla char *s and their ilk get caught on NULLs.
// This could definitely be a better string implementation,
// but it'll work for now...

#include <ctype.h>
#include <string.h>
#include "refcount.h"
#include "iterator.h"
#include "hashtable.h"
#include "algo.h"
#include "ptr.h"

class String {
public:
  typedef char value_type;
  typedef size_t hash_type;
  typedef ptrdiff_t difference_type;
  typedef size_t size_type;
  typedef value_type *iterator;
  typedef const value_type *const_iterator;

  // Sorry about the casts - I *really* want to
  // be using const_cast<> here, but that doesn't
  // seem to work with Solaris's CC... :-(
  String(const String &it) : data(it.data) {}
  String(const value_type *s, size_type n)
    : data(new Buf((value_type *)s, n, false)) {}
  String(const value_type *s, size_type n, hash_type hashval)
    : data(new Buf((value_type *)s, n, hashval, false)) {}
  String(const_iterator start, const_iterator finish)
    : data(new Buf((value_type *)start, finish - start, false)) {}
  String(const value_type *s)
    : data(new Buf((value_type *)s, false)) {}
  String() : data(new Buf()) {}
  String(String::size_type size) : data(new Buf(size)) {}
  String &operator=(const String &it) {
    data = it.data;
    return *this;
  }
  static String Copy(const value_type *s) {
    P<Buf> data = new Buf((value_type *)s, true);
    return String(data);
  }
  
  static String Copy(const value_type *s,
		     size_type n) {
    P<Buf> data = new Buf((value_type *)s, n, true);
    return String(data);
  }

  static String Copy(const String &it) {
    P<Buf> data = new Buf((value_type *)it.octets(), it.length(), true);
    return String(data);
  }
  
  String operator+(const String &it) const {
    P<Buf> dat = new Buf(length() + it.length());
    copy_n(octets(), length(), (value_type *)*dat);
    copy_n(it.octets(), it.length(), (value_type *)*dat + length());
    return String(dat);
  }
  
  String substr(const_iterator start,
		const_iterator finish) {
    return String(start, finish - start);
  }

  hash_type hash() const { return data->hash(); }
  hash_type rehash() const { return data->rehash(); }

private:

  // Grumble.  I hate compilers that don't bother to implement
  // the "mutable" keyword... not to mention const_cast<>()
  // (that's why we have the funny String * casts here.)
  void ensure_zero_termination() const {
    if ((*data)[data->size()] != '\0')
      ((String *)this)->data = new Buf(*data);
  }

public:

  String::value_type *c_str() {
    ensure_zero_termination();
    return *data;
  }
  const String::value_type *c_str() const {
    ensure_zero_termination();
    return *data;
  }
  String::value_type *octets() {
    return *data;
  }
  const String::value_type *octets() const {
    return *data;
  }
  iterator begin() { return data->begin(); }
  const_iterator begin() const { return data->begin(); }
  iterator end() { return data->end(); }
  const_iterator end() const { return data->end(); }
  size_type size() const { return data->size(); }
  size_type length() const { return data->size(); }
  friend ostream &operator<<(ostream &s, const String &str) {
    return (s.write((char *)*(str.data), str.data->size()));
  }
  value_type &operator[](size_type offset) {
    return (*data)[offset];
  }
  value_type operator[](size_type offset) const {
    return (*data)[offset];
  }
  bool operator==(const String &str) const {
    return (data->hash() == str.data->hash()
	    && data->size() == str.data->size()
	    && ! lexicographical_compare_3way(begin(), end(),
					      str.begin(), str.end()));
  }
  bool operator==(const char *str) const {
    return ! strcmp(octets(), str);
  }
  bool operator!=(const char *str) const {
    return strcmp(octets(), str);
  }
  int compare(const String &str) const {
    return lexicographical_compare_3way(begin(), end(),
					str.begin(), str.end());
  }
  bool operator<(const String &str) const {
    return lexicographical_compare(begin(), end(),
				   str.begin(), str.end());
  }
  void toLower() {
    // This should even work on UTF-8, assuming that
    // tolower will map 8bit characters appropriately.
    for (iterator it = begin(); it != end(); ++it)
      *it = tolower((unsigned char)*it);
  }

  class Buf : public Refcounted {
  public:
    Buf() : deleteit(false), hashvalue(0), length(0), data("") {}
    Buf(String::value_type *v, bool docopy)
      : deleteit(docopy),
	hashvalue(0),
	length(find(v, (String::value_type *)NULL, '\0') - v),
	data(docopy ? new String::value_type[length + 1] : v)
    {
      if (docopy)
	copy_n(v, length + 1, data);
    }
    Buf(String::value_type *v,
	String::size_type size, bool docopy)
      : deleteit(docopy),
	hashvalue(0),
	length(size),
	data(docopy ? new String::value_type[length + 1] : v)
    {
      if (docopy) {
	copy_n(v, length, data);
	data[length] = '\0';
      }
    }
    Buf(String::value_type *v,
	String::size_type size,
	String::hash_type hashval,
	bool docopy)
      : deleteit(docopy),
	hashvalue(hashval),
	length(size),
	data(docopy ? new String::value_type[length + 1] : v)
    {
      if (docopy) {
	copy_n(v, length, data);
	data[length] = '\0';
      }
    }
    Buf(const Buf &it)
      : deleteit(true),
	hashvalue(it.hashvalue),
	length(it.length),
	data(new String::value_type[length + 1])
    {
      copy_n(it.data, length, data);
      data[length] = '\0';
    }
    Buf(String::size_type size)
      : deleteit(true),
	hashvalue(0),
	length(size),
	data(new String::value_type[size + 1]) {
	  data[length] = '\0';
    }
    virtual ~Buf();
    String::hash_type rehash();	// forces a rehash
    String::hash_type hash() {
      if (hashvalue) return hashvalue;
      else return rehash();
    }
    String::size_type size() { return length; }
    String::iterator begin() { return data; }
    String::iterator end() { return data + length; }
    String::value_type &operator[](String::size_type offset) {
      return data[offset];
    }
    operator String::value_type *() { return data; }

  protected:
    mutable hash_type hashvalue;

  private:
    bool deleteit;
    String::size_type length;
    String::value_type *data;
  };

  String(P<Buf> &buf) : data(buf) {}

private:
  P<Buf> data;
};

//"So would you say that you've accumulated enough rope to hang
// yourself with?"  -- Dan Martinez <dmartine@Adobe.COM>

class StringTokenizer
  : public input_iterator<String, String::difference_type> {
public:
    StringTokenizer() : str(), first(0), last(0) {}
    StringTokenizer(const String &_str, String::value_type _c)
      : str(_str),
	c(_c),
	first(str.begin()),
	last(find(str.begin(), str.end(), c)) {}
    StringTokenizer(const StringTokenizer &s)
      : str(s.str), c(s.c), first(s.first), last(s.last) {}
    StringTokenizer &operator=(const StringTokenizer &s) {
      str = s.str;
      c = s.c;
      first = s.first;
      last = s.last;
      return *this;
    }
    StringTokenizer &operator++() {
      incr();
      return *this;
    }
    StringTokenizer operator++(int) {
      StringTokenizer tmp = *this;
      incr();
      return tmp;
    }
    String operator*() { return str.substr(first, last); }
    bool Done() const { return (first == str.end()); }
    bool Last() const { return (last == str.end()); }
private:
    void incr(void) {
      if (last != str.end()) last++;
      first = last;
      last = find(first, str.end(), c);
    }
    String str;
    String::value_type c;
    String::iterator first, last;
};

struct hash<String> {
  size_t operator()(const String &str) const {
    return str.hash();
  }
};

#endif // ACAP_STR_H
