/* timestamp.cc - timestamp implementation
 * Rob Earhart
 * $Id: timestamp.cc,v 1.4 1998/03/09 00:58:54 rob Exp $
 */
/***********************************************************
        Copyright 1997 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of CMU not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.
******************************************************************/

#include <limits.h>
#include <iostream.h>
#include <strstream.h>
#include <iomanip.h>
#include <stdio.h>
#include "timestamp.h"
#include "base64.h"
#include "response.h"
#include "ps.h"

TimestampFactory *TimestampFactory::instance = NULL;

pthread_once_t TimestampFactory::instance_init = PTHREAD_ONCE_INIT;

void TimestampFactory::Initialize() {
  instance = new TimestampFactory();
}

TimestampFactory *TimestampFactory::Instance() {
  if (! instance)
    pthread_once(&instance_init, &Initialize);
  return instance;
}

Ptr<Timestamp> TimestampFactory::NewStamp() {
  time_t now = time(NULL);
  unsigned long frac;

  Grab g(factory_mutex);
  if (now > last.sec) {
    last.sec = now;
    last.frac = 0;
  } else {
    if (last.frac == ULONG_MAX) {
      // Rollover - shouldn't ever happen, but it might if we somehow
      // manage to start ULONG_MAX transactions within a single second.
      // If it does happen, we'll just increment to the next second
      // and keep going; eventually, hopefully, the load will lighten
      // up a little, and the real clock and our internal "now" will
      // sync up again.
      now = ++last.sec;
      last.frac = 0;
    } else {
      now = last.sec;
      last.frac++;
    }
  }
  frac = last.frac;
  return new Timestamp(now, frac);
}

void TimestampFactory::RemoveStamp(Timestamp *t) {
  // We need to ensure that none of the blobs used by this
  // timestamp are free'd until all previous timestamps
  // have completed; the easiest way is to hand the next-previous
  // timestamp all the blobs in this timestamp.
  Grab g(factory_mutex);
  stamps_t::iterator it = stamps.find(t);
  if (it != stamps.begin()) {
    stamps_t::iterator prev = it;
    --prev;
    Grab p((*prev)->blobs_mutex);
    (*prev)->blobs.insert((*it)->blobs.begin(), (*it)->blobs.end());
  }
}

APtr<char> Timestamp::MakeLongStr(const Stamp &stamp) {
  // This is a *little* twisted;
  // the net effect is to set the char array
  // ulongmax to the string form of ULONG_MAX,
  // so that at runtime we can take the sizeof(ulongmax)
  // and have the length of the string needed to hold
  // the value of ULONG_MAX.

  // Then again, this whole routine is pretty twisted.
#define STRINGIFY1(x) #x
#define STRINGIFY(x) STRINGIFY1(x)
  static char ulongmax[] = STRINGIFY(ULONG_MAX);
#undef STRINGIFY1
#undef STRINGIFY

  char fracb[sizeof(ulongmax)];	// definitely big enough

  // The '3' here takes care of the NULL at the end of
  // ulongmax, as well as the trailing UL.
  sprintf(fracb, "%0*ld", sizeof(ulongmax) - 3, stamp.frac);

  struct tm gmt;
  gmtime_r(&stamp.sec, &gmt);
  // 14 for the date
  // fracb for the ulong
  // 1 for the null
  // There're probably better ways to do this,
  // but this works...
  APtr<char> result = new char[15 + strlen(fracb)];
  sprintf(result, "%04d%02d%02d%02d%02d%02d%s",
	  gmt.tm_year + 1900,
	  gmt.tm_mon + 1,
	  gmt.tm_mday,
	  gmt.tm_hour,
	  gmt.tm_min,
	  gmt.tm_sec,
	  fracb);
  return result;
}

APtr<char> Timestamp::MakeShortStr(const Stamp &stamp) {
  // All we do here is compress sec.frac into the shortest possible
  // string that doesn't contain anything interesting (newlines,
  // spaces, quotes, etc), without sacrificing any bits, and hopefully
  // maintaining ordering under octet comparison.  Base64 turns out to
  // be perfect for this...
  APtr<char> result = Base64::ToBase64((const char *)&stamp, sizeof(stamp));
  return result;
}

Timestamp::Stamp::Stamp(String &str) {
  struct tm gmt;
  char c;
  int n_scanned = sscanf(str.c_str(), "%4u%2u%2u%2u%2u%2u%u%c",
			 &gmt.tm_year,
			 &gmt.tm_mon,
			 &gmt.tm_mday,
			 &gmt.tm_hour,
			 &gmt.tm_min,
			 &gmt.tm_sec,
			 &frac,
			 &c /* we expect to not be able to read this char */);
  if (n_scanned < 6 || n_scanned > 7)
    throw Response::BAD("Bad modtime");

  if (n_scanned == 6)
    frac = 0;

  gmt.tm_year-=1900;
  gmt.tm_mon--;

  sec = mktime(&gmt);
}

Timestamp::~Timestamp() {
  TimestampFactory::Instance()->RemoveStamp(this);
}

void Timestamp::AddPSBlob(PSBlob *blob) {
  Grab g(blobs_mutex);
  blobs.insert(blob);
}
