/* trx.cc - transaction implementation
 * Rob Earhart
 * $Id: trx.cc,v 1.2 1998/03/09 00:58:56 rob Exp $
 */
/***********************************************************
        Copyright 1998 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of CMU not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.
******************************************************************/

#include <unistd.h>
#include <limits.h>
#include "trx.h"
#include "unixerror.h"
#include "ps.h"
#include "util.h"

void Trx::Recover() {
}

void Trx::WriteSetPS(PS *ps) {
  pset.insert(ps->GetResources());
  tfile << 'p' << ps->Path() << '\0';
}

Trx::Trx(Timestamp *t) : current_ps(NULL), committed(false), tstamp(t) {}

void Trx::Commit() {
  committed = true;
}

static void Corrupt() {
  DEBUG("Trx: Corruption detected");
  throw InternalError("Corrupted transaction file discovered");
}

void Trx::Apply(istream &trx) {
  char c;
  char zero;
  FD fd;
  //  DEBUG("Trx: Processing");
  String current_path;
  while (trx.get(c))
    switch (c) {
    case 'p': {
      // SetPS
      // DEBUG("Trx: Setting PS");
      char *path = NULL;
      try {
	path = new char[PATH_MAX];
	trx.get(path, PATH_MAX, '\0');
	trx.get(zero);
	if (! trx || zero != '\0') Corrupt();
	if (current_path != path) {
	  // DEBUG("Trx: Setting PS to " << path);
	  if (fd != -1 && fsync(fd))
	    throw UnixError("Synchronizing dataset");
	  fd.Open(path, O_RDWR);
	  current_path = String::Copy(path);
	} else {
	  // DEBUG("Trx: PS already set to " << path);
	}
      } catch (...) {
	if (path) delete [] path;
	throw;
      }
      delete [] path;
      // DEBUG("Trx: PS Set");
    }
    break;
    case 'g': {
      // DEBUG("Trx: Growing to offset");
      // Grow to offset
      size_t size;
      (trx >> size).get(zero);
      if (! trx || zero != '\0') Corrupt();
      // DEBUG("Trx: Growing to " << size);
      ftruncate(fd, size);
      // DEBUG("Trx: Grew");
    }
    break;
    case 'o': {
      // DEBUG("Trx: Writing offset");
      // Write offset at offset
      off_t offset;
      off_t data;
      (trx >> offset).get(zero);
      if (! trx || zero != '\0') Corrupt();
      (trx >> data).get(zero);
      if (! trx || zero != '\0') Corrupt();
      // DEBUG("Trx: Writing offset " << data << " at offset " << offset);
      if (pwrite(fd, &data, sizeof(data), offset) != sizeof(data))
	throw UnixError("Updating dataset");
      // DEBUG("Trx: Wrote offset");
    }
    break;
    case 'd': {
      // DEBUG("Trx: Writing data");
      // Write data
      off_t offset;
      size_t len;
      char *data = NULL;
      (trx >> offset).get(zero);
      if (! trx || zero != '\0') Corrupt();
      (trx >> len).get(zero);
      if (! trx || zero != '\0') Corrupt();
      if (len) {
	try {
	  data = new char[len];
	  trx.read(data, len).get(zero);
	  if (! trx || zero != '\0') Corrupt();
	  // DEBUG("Trx: Writing " << len << " bytes of data at " << offset);
	  if (pwrite(fd, data, len, offset) != len)
	    throw UnixError("Updating dataset");
	} catch (...) {
	  if (data) delete [] data;
	  throw;
	}
	delete [] data;
      } else {
	trx.get(zero);
	if (! trx || zero != '\0') Corrupt();
      }
      // DEBUG("Trx: Wrote data");
    }
    break;
    case 'a':
      {
	// DEBUG("Trx: Allocing data");
	// Alloc data
	off_t offset;
	size_t len;
	(trx >> offset).get(zero);
	if (! trx || zero != '\0') Corrupt();
	(trx >> len).get(zero);
	if (! trx || zero != '\0') Corrupt();
	// DEBUG("Trx: Recording allocation: " << len << " bytes at " << offset);
	// XXX
      }
    break;
    case 'f':
      {
	// DEBUG("Trx: Freeing data");
	// Free data
	off_t offset;
	size_t len;
	(trx >> offset).get(zero);
	if (! trx || zero != '\0') Corrupt();
	(trx >> len).get(zero);
	if (! trx || zero != '\0') Corrupt();
	// DEBUG("Trx: Recording deallocation: " << len << " bytes at " << offset);
	// XXX
      }
    break;
    default:
      DEBUG("Trx: Bad operand detected");
      Corrupt();
    }
  if (fd != -1 && fsync(fd))
    throw UnixError("Synchronizing dataset");
  // DEBUG("Trx: processed");
}

Trx::~Trx() {
  tfile.rdbuf()->freeze(1);
  tfile.seekg(0);
#ifndef NDEBUG
  tfile.seekg(0);
#endif
  Apply(tfile);
  tfile.rdbuf()->freeze(0);
}
